@@ -1,5 +1,5 @@
 /*
- * $Id: com-config.c,v 1.5 1999/09/23 18:34:44 wiegand Exp $
+ * $Id: com-config.c,v 1.6 2002/01/14 18:12:20 mt Exp $
  *
  * Common functions for configuration reading
  *
@@ -30,7 +30,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: com-config.c,v 1.5 1999/09/23 18:34:44 wiegand Exp $";
+static char rcsid[] = "$Id: com-config.c,v 1.6 2002/01/14 18:12:20 mt Exp $";
 #endif
 
 #include <config.h>
@@ -258,13 +258,9 @@ void config_read(char *file, int dflg)
 
 	if ((fp = fopen(file, "r")) == NULL) {
 		syslog_error("can't open config file '%.*s'",
-					MAX_PATH_SIZE, file);
+		             MAX_PATH_SIZE, file);
 		exit(EXIT_FAILURE);
 	}
-	if (dflg == 0) {
-		syslog_write(T_INF, "Config-File: '%.*s'",
-					MAX_PATH_SIZE, file);
-	}
 
 	/*
 	** Prepare the global section
@@ -398,12 +394,13 @@ void config_read(char *file, int dflg)
 	** Do we just want to validate the interpretation?
 	*/
 	if (dflg != 0) {
+		printf("Config-File: '%.*s'\n", MAX_PATH_SIZE, file);
 		for (sect = sechead; sect; sect = sect->next) {
 			printf("Config-Section ------ '%.*s'\n",
 				MAX_CONF_NAME,
 				sect->name ? sect->name : "(-global-)");
 			for (conf = sect->conf; conf; conf = conf->next) {
-				printf("Config: %-*.*s = '%.*s'\n",
+				printf("Config:        %-*.*s = '%.*s'\n",
 					MIN_CONF_NAME, MAX_CONF_NAME,
 					conf->name,
 					MAX_PATH_SIZE, conf->data);
@@ -415,6 +412,8 @@ void config_read(char *file, int dflg)
 	/*
 	** Inform the possible auditor
 	*/
+	syslog_write(T_INF, "Config-File: '%.*s'",
+			MAX_PATH_SIZE, file);
 	for (sect = sechead; sect; sect = sect->next) {
 		syslog_write(T_INF, "Config-Section ------ '%.*s'",
 				MAX_CONF_NAME,
@@ -428,6 +427,27 @@ void config_read(char *file, int dflg)
 	}
 }
 
+void config_dump(FILE *fd)
+{
+	SECTION *sect;
+	CONFIG *conf;
+
+	if(NULL == fd)
+		return;
+
+	for (sect = sechead; sect; sect = sect->next) {
+		fprintf(fd, "[%.*s]\n", MAX_CONF_NAME,
+			sect->name ? sect->name : "-Global-");
+
+		for (conf = sect->conf; conf; conf = conf->next) {
+			fprintf(fd, "%-*.*s %.*s\n",
+				MIN_CONF_NAME, MAX_CONF_NAME,
+				conf->name,
+				MAX_PATH_SIZE, conf->data);
+		}
+		fprintf(fd, "\n");
+	}
+}
 
 /* ------------------------------------------------------------ **
 **
@@ -968,6 +988,9 @@ gid_t config_gid(char *snam, char *name, gid_t dflt)
 
 /* ------------------------------------------------------------
  * $Log: com-config.c,v $
+ * Revision 1.6  2002/01/14 18:12:20  mt
+ * implemented config_dump function to dump in-memory config to a FILE stream
+ *
  * Revision 1.5  1999/09/23 18:34:44  wiegand
  * remove white space at line start (incl. continuation lines)
  *
