@@ -1,5 +1,5 @@
 /*
- * $Id: com-socket.c,v 1.4 1999/09/21 05:42:28 wiegand Exp $
+ * $Id: com-socket.c,v 1.6 2002/01/14 18:26:55 mt Exp $
  *
  * Common functions for TCP/IP sockets
  *
@@ -30,7 +30,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: com-socket.c,v 1.4 1999/09/21 05:42:28 wiegand Exp $";
+static char rcsid[] = "$Id: com-socket.c,v 1.6 2002/01/14 18:26:55 mt Exp $";
 #endif
 
 #include <config.h>
@@ -69,25 +69,64 @@ static char rcsid[] = "$Id: com-socket.c,v 1.4 1999/09/21 05:42:28 wiegand Exp $
 #endif
 
 #include <sys/ioctl.h>
+#if defined(HAVE_SYS_FILIO_H)
+#include <sys/filio.h>
+#endif
 
-#include <netdb.h>
-#include <netinet/ip.h>
-#include <netinet/in.h>
 #include <sys/socket.h>
+#include <netinet/in.h>
 #include <arpa/inet.h>
+#include <netdb.h>
 
+#if defined(HAVE_STROPTS_H)
+#   include <stropts.h>
+#endif
+#if defined(HAVE_SYS_CONF_H)
+#   include <sys/conf.h>
+#endif
 #if defined(HAVE_SYS_SOCKIO_H)
 #  include <sys/sockio.h>
 #endif
+#if defined(HAVE_NETINET_IN_SYSTM_H)
+#   include <netinet/in_systm.h>
+#endif
+#if defined (SOLARIS)
+#   define FIONREAD I_NREAD
+#endif
+
+#if defined(HAVE_NET_IF_H)
+#  include <net/if.h>
+#endif
+
+#if defined(HAVE_NETINET_IP_H)
+#   include <netinet/ip.h>
+#endif
+
+#if defined(HAVE_NETINET_IP_COMPAT_H)
+#   include <netinet/ip_compat.h>
+#endif
+#if defined(HAVE_NETINET_IP_FIL_COMPAT_H)
+#   include <netinet/ip_fil_compat.h>
+#endif
+#if defined(HAVE_NETINET_IP_FIL_H)
+#   include <netinet/ip_fil.h>
+#endif
+#if defined(HAVE_NETINET_IP_NAT_H)
+#   include <netinet/ip_nat.h>
+#endif
+
+#if defined(HAVE_LINUX_NETFILTER_IPV4_H)
+#   include <linux/netfilter_ipv4.h>
+#endif
 
 #if defined(HAVE_LIBWRAP)
-#  if defined(HAVE_SYSLOG_H)
-#    include <syslog.h>
-#  endif
-#  if defined(NEED_SYS_SYSLOG_H)
-#    include <sys/syslog.h>
-#  endif
-#  include <tcpd.h>
+#   if defined(HAVE_SYSLOG_H)
+#      include <syslog.h>
+#   endif
+#   if defined(NEED_SYS_SYSLOG_H)
+#      include <sys/syslog.h>
+#   endif
+#   include <tcpd.h>
 #endif
 
 #include "com-config.h"
@@ -103,6 +142,10 @@ static char rcsid[] = "$Id: com-socket.c,v 1.4 1999/09/21 05:42:28 wiegand Exp $
 #  define SOMAXCONN	5	/* Default accept queue size	*/
 #endif
 
+#if !defined(NETSIZ)
+#  define NETSIZ	8192	/* Default network buffer size	*/
+#endif
+
 
 /* ------------------------------------------------------------ */
 
@@ -127,6 +170,7 @@ int allow_severity = LOG_INFO;	/* TCP Wrapper log levels	*/
 int deny_severity  = LOG_WARNING;
 #endif
 
+static int maxrecv_bufsiz = -1;	/* max receive buffer size	*/
 
 /* ------------------------------------------------------------ **
 **
@@ -170,6 +214,16 @@ int socket_listen(u_int32_t addr, u_int16_t port, ACPT_CB func)
 	if (initflag == 0) {
 		atexit(socket_cleanup);
 		initflag = 1;
+
+		/*
+		** Check if we should limit the recv buffer size...
+		** (because the link on the write side is much slower)
+		*/
+		if(maxrecv_bufsiz < 0) {
+			maxrecv_bufsiz = config_int(NULL, "MaxRecvBufSize", 0);
+			if(maxrecv_bufsiz < 0)
+				maxrecv_bufsiz = 0;
+		}
 	}
 
 	/*
@@ -248,7 +302,8 @@ void socket_lclose(int shut)
 
 static void socket_accept(void)
 {
-	char peer[PEER_LEN];
+	char peer[PEER_LEN] = {0};
+	char dest[PEER_LEN] = {0};
 	struct sockaddr_in saddr;
 	int nsock, len;
 
@@ -262,10 +317,15 @@ static void socket_accept(void)
 		syslog_error("can't accept client");
 		return;
 	}
-	strcpy(peer, inet_ntoa(saddr.sin_addr));
+
+	misc_strncpy(peer, inet_ntoa(saddr.sin_addr), sizeof(peer));
+	memset(&saddr, 0, sizeof(saddr));
+	if( !getsockname(nsock, (struct sockaddr *)&saddr, &len)) {
+		misc_strncpy(dest, inet_ntoa(saddr.sin_addr), sizeof(dest));
+	}
 
 #if defined(COMPILE_DEBUG)
-	debug(2, "accepted %d=%s", nsock, peer);
+	debug(2, "accepted %d=%s on %s", nsock, NIL(peer), NIL(dest));
 #endif
 
 #if defined(HAVE_LIBWRAP)
@@ -274,14 +334,18 @@ static void socket_accept(void)
 	*/
 	if (config_bool(NULL, "TCPWrapper", 0)) {
 		struct request_info req;
+		char *wn;
 
-		request_init(&req, RQ_DAEMON, misc_getprog(),
+		wn = config_str(NULL, "TCPWrapperName", misc_getprog());
+		if( !(wn && *wn)) wn = "ftp-proxy"; /* fall back... */
+
+		request_init(&req, RQ_DAEMON, wn,
 					RQ_FILE, nsock, NULL);
 		fromhost(&req);
 		if (hosts_access(&req) == 0) {
 			close(nsock);
 			syslog_write(U_ERR,
-				"reject: '%s' (Wrap)", peer);
+				"%s reject: '%s' (Wrap)", wn, peer);
 			return;
 		}
 	}
@@ -322,6 +386,16 @@ HLS *socket_init(int sock)
 	if (initflag == 0) {
 		atexit(socket_cleanup);
 		initflag = 1;
+
+		/*
+		** Check if we should limit the recv buffer size...
+		** (because the link on the write side is much slower)
+		*/
+		if(maxrecv_bufsiz < 0) {
+			maxrecv_bufsiz = config_int(NULL, "MaxRecvBufSize", 0);
+			if(maxrecv_bufsiz < 0)
+				maxrecv_bufsiz = 0;
+		}
 	}
 
 	hls = (HLS *) misc_alloc(FL, sizeof(HLS));
@@ -329,9 +403,9 @@ HLS *socket_init(int sock)
 	hlshead   = hls;
 
 	if ((hls->sock = sock) != -1) {
-		hls->addr = socket_sck2addr(sock,
-					REM_END, &(hls->port));
-		strcpy(hls->peer, socket_addr2str(hls->addr));
+		hls->addr = socket_sck2addr(sock, REM_END, &(hls->port));
+		misc_strncpy(hls->peer, socket_addr2str(hls->addr),
+		             sizeof(hls->peer));
 	} else {
 		hls->addr = 0;
 		hls->port = 0;
@@ -340,11 +414,15 @@ HLS *socket_init(int sock)
 
 	hls->kill = 0;
 	hls->flag = 0;
+	hls->more = 0;
 	hls->ctyp = "???-????";
 
 	hls->wbuf = NULL;
 	hls->rbuf = NULL;
 
+	hls->wcnt = 0;
+	hls->rcnt = 0;
+
 #if defined(COMPILE_DEBUG)
 	debug(2, "created HLS for %d=%s:%d",
 			hls->sock, hls->peer, (int) hls->port);
@@ -508,11 +586,30 @@ char *socket_gets(HLS *hls, char *ptr, int len)
 	/*
 	** Transfer at most one line of data
 	*/
+	hls->more = 0;
 	for (buf = hls->rbuf, cnt = 0; buf != NULL && cnt < len; ) {
 		if (buf->cur >= buf->len) {
 			hls->rbuf = buf->next;
 			misc_free(FL, buf);
-			buf = hls->rbuf;
+			if(NULL == (buf = hls->rbuf)) {
+				/*
+				** last buffer in HLS and no EOL found;
+				** restore the data into the HLS and
+				** exit to wait to read more
+				*/
+				hls->more = 1;
+				hls->rbuf = (BUF *)misc_alloc(FL, sizeof(BUF) + cnt);
+				hls->rbuf->len = cnt;
+				hls->rbuf->cur = 0;
+				memcpy(hls->rbuf->dat, ptr, cnt);
+#if defined(COMPILE_DEBUG)
+				debug(4, "preread %d bytes while waiting "
+					 "for end-of-line: '%.128s'%s",
+					 cnt, hls->rbuf->dat,
+					 (cnt > 128) ? "..." : "");
+#endif
+				return NULL;
+			}
 			continue;
 		}
 		if (buf->dat[buf->cur] == '\r')
@@ -647,7 +744,7 @@ int socket_write(HLS *hls, char *ptr, int len)
 int socket_printf(HLS *hls, char *fmt, ...)
 {
 	va_list aptr;
-	char str[8192];
+	char str[NETSIZ];
 	int len;
 	BUF *buf, *tmp;
 
@@ -660,8 +757,13 @@ int socket_printf(HLS *hls, char *fmt, ...)
 	/*
 	** Prepare the new stuff to be written
 	*/
+	memset(str, 0, sizeof(str));
 	va_start(aptr, fmt);
+#if defined(HAVE_VSNPRINTF)
+	vsnprintf(str, sizeof(str), fmt, aptr);
+#else
 	vsprintf(str, fmt, aptr);
+#endif
 	va_end(aptr);
 	len = strlen(str);
 
@@ -811,10 +913,12 @@ int socket_exec(int timeout, int *close_flag)
 			debug(4, "FD_SET %s for W", hls->ctyp);
 #endif
 		}
-		FD_SET(hls->sock, &rfds);
+		if(hls->more >= 0) {
+			FD_SET(hls->sock, &rfds);
 #if defined(COMPILE_DEBUG)
-		debug(4, "FD_SET %s for R", hls->ctyp);
+			debug(4, "FD_SET %s for R", hls->ctyp);
 #endif
+		}
 	}
 
 	/*
@@ -851,16 +955,20 @@ int socket_exec(int timeout, int *close_flag)
 	if (lsock != -1 && FD_ISSET(lsock, &rfds))
 		socket_accept();
 	for (hls = hlshead; hls != NULL; hls = hls->next) {
+
 		if (hls->sock == -1)
 			continue;
-		if (FD_ISSET(hls->sock, &rfds))
-			socket_ll_read(hls);
-		if (hls->sock == -1)	/* May be dead by now */
-			continue;
+
 		if (FD_ISSET(hls->sock, &wfds))
 			socket_ll_write(hls);
 		if (hls->sock == -1)	/* May be dead by now */
 			continue;
+
+		if (FD_ISSET(hls->sock, &rfds))
+			socket_ll_read(hls);
+		if (hls->sock == -1)	/* May be dead by now */
+			continue;
+
 		if (hls->kill != 0 && hls->wbuf == NULL) {
 			close(hls->sock);
 			hls->sock = -1;
@@ -917,9 +1025,9 @@ static void socket_ll_read(HLS *hls)
 		shutdown(hls->sock, 2);		/* the "acceptor" */
 		close(hls->sock);
 		hls->sock = nsock;
-		hls->addr = socket_sck2addr(nsock,
-					REM_END, &(hls->port));
-		strcpy(hls->peer, socket_addr2str(hls->addr));
+		hls->addr = socket_sck2addr(nsock, REM_END, &(hls->port));
+		misc_strncpy(hls->peer, socket_addr2str(hls->addr),
+		             sizeof(hls->peer));
 #if defined(COMPILE_DEBUG)
 		debug(2, "accept %s (%d) from %s",
 				hls->ctyp, hls->sock, hls->peer);
@@ -949,6 +1057,16 @@ static void socket_ll_read(HLS *hls)
 #endif
 		return;
 	}
+#if defined(COMPILE_DEBUG)
+	debug(4, "ll_read: FIONREAD reported %d bytes for %s",
+		len, hls->ctyp);
+#endif
+
+	/*
+	** Limit the receive buffer sizes
+	*/
+	if(maxrecv_bufsiz > 0 && len > maxrecv_bufsiz)
+		len = maxrecv_bufsiz;
 
 	/*
 	** Now read the data that is waiting
@@ -970,6 +1088,11 @@ static void socket_ll_read(HLS *hls)
 	buf->cur = 0;
 	buf->flg = 0;
 
+	/*
+	** Update byte conter
+	*/
+	hls->rcnt += len;
+
 	/*
 	** Chain the newly filled buffer
 	*/
@@ -983,8 +1106,9 @@ static void socket_ll_read(HLS *hls)
 	buf->next = NULL;
 
 #if defined(COMPILE_DEBUG)
-	debug(3, "ll_read %s %d=%s: %d bytes",
-			hls->ctyp, hls->sock, hls->peer, len);
+	debug(3, "ll_read %s %d=%s: %d/%d bytes",
+			hls->ctyp, hls->sock, hls->peer,
+			len, hls->rcnt);
 #endif
 }
 
@@ -1032,10 +1156,20 @@ static void socket_ll_write(HLS *hls)
 			break;	/* At least the first write was ok */
 		}
 
+#if defined(COMPILE_DEBUG)
+		debug(4, "ll_write %s %d=%s: sent %d bytes",
+			hls->ctyp, hls->sock, hls->peer, cnt);
+#endif
+
+		/*
+		** Update byte conter
+		*/
+		tot += cnt;
+		hls->wcnt += cnt;
+
 		/*
 		** Advance the write pointers
 		*/
-		tot += cnt;
 		if ((buf->cur += cnt) < buf->len)
 			break;	/* Partly sent, try again later */
 
@@ -1048,8 +1182,9 @@ static void socket_ll_write(HLS *hls)
 	}
 
 #if defined(COMPILE_DEBUG)
-	debug(3, "ll_write %s %d=%s: %d bytes",
-			hls->ctyp, hls->sock, hls->peer, tot);
+	debug(3, "ll_write %s %d=%s: %d/%d bytes",
+			hls->ctyp, hls->sock, hls->peer,
+			tot, hls->wcnt);
 #endif
 }
 
@@ -1095,41 +1230,60 @@ char *socket_msgline(char *fmt)
 		switch (*++fmt) {
 			case 'b':
 			case 'B':
-				strcpy(tmp, misc_getdate());
+				strncpy(tmp, misc_getdate(), sizeof(tmp)-1);
 				break;
 			case 'd':
 			case 'D':
+#if defined(HAVE_SNPRINTF)
+				snprintf(tmp, sizeof(tmp),
+						"%04d/%02d/%02d",
+						t->tm_year + 1900,
+						t->tm_mon  + 1,
+						t->tm_mday);
+#else
 				sprintf(tmp, "%04d/%02d/%02d",
 						t->tm_year + 1900,
 						t->tm_mon  + 1,
 						t->tm_mday);
+#endif
 				break;
 			case 'h':
 			case 'H':
 				if (gethostname(tmp, sizeof(tmp)) < 0)
-					strcpy(tmp, "[unknown host]");
+					strncpy(tmp, "[unknown host]",
+						sizeof(tmp)-1);
 				break;
 			case 'n':
 			case 'N':
-				if (getdomainname(tmp, sizeof(tmp)) < 0)
-					strcpy(tmp, "[unknown network]");
+				if (getfqdomainname(tmp, sizeof(tmp)) < 0)
+					strncpy(tmp, "[unknown network]",
+						sizeof(tmp)-1);
 				break;
 			case 't':
 			case 'T':
+#if defined(HAVE_SNPRINTF)
+				snprintf(tmp, sizeof(tmp),
+						"%02d:%02d:%02d",
+						t->tm_hour,
+						t->tm_min,
+						t->tm_sec);
+#else
 				sprintf(tmp, "%02d:%02d:%02d",
 						t->tm_hour,
 						t->tm_min,
 						t->tm_sec);
+#endif
 				break;
 			case 'v':
 			case 'V':
-				strcpy(tmp, misc_getvers());
+				strncpy(tmp, misc_getvers(), sizeof(tmp)-1);
 				break;
 			case '%':
 				tmp[0] = '%';
 				break;
 			default:
 		}
+		tmp[sizeof(tmp)-1] = '\0'; /* paranoia :-) */
 		j = strlen(tmp);
 		if ((i + j) < (sizeof(str) - 64)) {
 			memcpy(str + i, tmp, j);
@@ -1141,6 +1295,147 @@ char *socket_msgline(char *fmt)
 }
 
 
+/* ------------------------------------------------------------ **
+**
+**	Function......:	socket_d_bind
+**
+**	Parameters....:	sock		socket descriptor
+**			addr		IP address we want to bind
+**			lrng		Lower TCP port range limit
+**			urng		Upper TCP port range limit
+**			incr		use rand or increment mode
+**
+**	Return........:	bound port, 0 (INPORT_ANY) on failure
+**
+**	Purpose.......: Binds a socket, taking care of a given
+**                      port range using rand or incrementing
+**                      the port number.
+**                      Note: this function covers also dynamic
+**                            ports assigning with a 0 range:
+**                            lrng = urng = 0 ( = INPORT_ANY)
+**
+** ------------------------------------------------------------ */
+u_int16_t socket_d_bind(int sock, u_int32_t addr,
+			u_int16_t lrng, u_int16_t urng,
+			int incr)
+{
+	struct sockaddr_in saddr;
+	u_int16_t          port = INPORT_ANY;
+	int                retry= MAX_RETRIES, err = -1;
+
+	/* Sanity check */
+	if(sock < 0)
+		return INPORT_ANY;
+
+	/* check port range */
+	if( !(lrng<=urng)) {
+#if defined(COMPILE_DEBUG)
+		debug(2, "socket_d_bind: invalid port range %d-%d",
+		         lrng, urng);
+#endif
+		return INPORT_ANY;
+	}
+
+	memset(&saddr, 0, sizeof(saddr));
+	saddr.sin_addr.s_addr = htonl(addr);
+	saddr.sin_family      = AF_INET;
+
+#if defined(COMPILE_DEBUG)
+	debug(4, "socket_d_bind using %s", incr ? "increment" : "random");
+#endif
+
+	if(incr) {
+		for (port = lrng; err && (port <= urng); port++) {
+
+			saddr.sin_port = htons(port);
+
+			while(0 <= retry--) {
+				errno = 0;
+				err = bind(sock, (struct sockaddr *)&saddr,
+				                 sizeof(saddr));
+#if defined(COMPILE_DEBUG)
+				debug(2, "bind %s:%d result: %d, status: %s",
+				         socket_addr2str(addr),
+				         port, err, strerror(errno));
+#endif
+				if(0 == err) {
+					/* bind succeed */
+#if defined(COMPILE_DEBUG)
+					debug(2, "bind succeeded,"
+						 "port: %d, result: %d",
+						 port, err);
+#endif
+					retry = -1; break;
+				} else {
+					/* bind failed: fatal error? */
+					if( !(EINTR == errno ||
+					      EAGAIN == errno ||
+					      EADDRINUSE == errno)) {
+#if defined(COMPILE_DEBUG)
+						debug(4, "bind failed,"
+						      "result: %d, error %s",
+						      err, strerror(errno));
+#endif
+						return INPORT_ANY;
+					}
+				}
+			}
+		}
+	} else {
+		int port_range = (urng - lrng) + 1;
+
+		while(err && (0 < port_range--)) {
+
+			port = misc_rand(lrng, urng);
+			saddr.sin_port = htons(port);
+
+			while(0 <= retry--) {
+				err = bind(sock, (struct sockaddr *)&saddr,
+				                 sizeof(saddr));
+#if defined(COMPILE_DEBUG)
+				debug(2, "bind %s:%d, result: %d, status: %s",
+				         socket_addr2str(addr),
+				         port, err, strerror(errno));
+#endif
+				if(0 == err) {
+					/* bind succeed */
+#if defined(COMPILE_DEBUG)
+					debug(2, "bind succeeded, port: %d,"
+						 "result: %d", port, err);
+#endif
+					retry = -1; break;
+				} else {
+					/* bind failed: fatal error? */
+					if( !(EINTR == errno ||
+					      EAGAIN == errno ||
+					      EADDRINUSE == errno)) {
+#if defined(COMPILE_DEBUG)
+						debug(2, "bind failed, "
+							"result: %d, error %s",
+							err, strerror(errno));
+#endif
+						return INPORT_ANY;
+					}
+				}
+			}
+		}
+	}
+
+	if((0 == err) &&
+	   (INADDR_NONE != socket_sck2addr(sock, LOC_END, &port)))
+	{
+#if defined(COMPILE_DEBUG)
+		debug(2, "bound socket to port %d", port);
+#endif
+		return port;
+	}
+#if defined(COMPILE_DEBUG)
+	debug(2, "bind error - port %d, IP %d",
+		port, socket_sck2addr(sock, LOC_END, NULL));
+#endif
+	return INPORT_ANY;
+}
+
 /* ------------------------------------------------------------ **
 **
 **	Function......:	socket_d_listen
@@ -1150,6 +1445,7 @@ char *socket_msgline(char *fmt)
 **			urng		Upper TCP port range limit
 **			phls		Pointer where HLS will go
 **			ctyp		Desired comms type identifier
+**			incr		use rand or incremental bind
 **
 **	Return........:	Listening port (0=failure)
 **
@@ -1158,11 +1454,12 @@ char *socket_msgline(char *fmt)
 **
 ** ------------------------------------------------------------ */
 
-u_int16_t socket_d_listen(u_int32_t addr, u_int16_t lrng,
-			u_int16_t urng, HLS **phls, char *ctyp)
+u_int16_t socket_d_listen(u_int32_t addr,
+			  u_int16_t lrng, u_int16_t urng,
+			  HLS **phls, char *ctyp,
+			  int incr)
 {
-	struct sockaddr_in saddr;
-	int sock;
+	int       sock;
 	u_int16_t port;
 
 	if (phls == NULL || ctyp == NULL)	/* Sanity check	*/
@@ -1177,23 +1474,9 @@ u_int16_t socket_d_listen(u_int32_t addr, u_int16_t lrng,
 	}
 	socket_opts(sock, SK_LISTEN);
 
-	memset(&saddr, 0, sizeof(saddr));
-	saddr.sin_addr.s_addr = htonl(addr);
-	saddr.sin_family      = AF_INET;
-
-	/*
-	** Bind the socket, taking care of a given port range.
-	** Note: this function covers also lrng = urng = 0 ...
-	*/
-	for (port = lrng; port <= urng; port++) {
-		saddr.sin_port = htons(port);
-		if (bind(sock, (struct sockaddr *) &saddr,
-					sizeof(saddr)) == 0)
-			break;
-		if (errno != EADDRINUSE)
-			port = urng + 1;
-	}
-	if (port > urng) {		/* nothing found?	*/
+	port = socket_d_bind(sock, addr, lrng, urng, incr);
+	if (INPORT_ANY == port) {
+		/* nothing found? */
 		close(sock);
 		return 0;
 	}
@@ -1207,10 +1490,9 @@ u_int16_t socket_d_listen(u_int32_t addr, u_int16_t lrng,
 	(*phls)->sock = sock;
 	(*phls)->ctyp = ctyp;
 
-	addr = socket_sck2addr(sock, LOC_END, &port);
 #if defined(COMPILE_DEBUG)
 	debug(2, "listen: %s (fd=%d) %s:%d", (*phls)->ctyp,
-		(*phls)->sock, socket_addr2str(addr), (int) port);
+		(*phls)->sock, socket_addr2str(addr), (int)port);
 #endif
 	return port;
 }
@@ -1227,6 +1509,7 @@ u_int16_t socket_d_listen(u_int32_t addr, u_int16_t lrng,
 **			urng		Upper TCP port range limit
 **			phls		Pointer where HLS will go
 **			ctyp		Desired comms type identifier
+**			incr		use rand or incremental bind
 **
 **	Return........:	Local end of connected port (0=failure)
 **
@@ -1236,61 +1519,114 @@ u_int16_t socket_d_listen(u_int32_t addr, u_int16_t lrng,
 ** ------------------------------------------------------------ */
 
 u_int16_t socket_d_connect(u_int32_t addr, u_int16_t port,
-			u_int32_t ladr, u_int16_t lrng,
-			u_int16_t urng, HLS **phls, char *ctyp)
+			   u_int32_t ladr,
+			   u_int16_t lrng, u_int16_t urng,
+			   HLS **phls, char *ctyp,
+			   int incr)
 {
 	struct sockaddr_in saddr;
-	int sock;
-	u_int16_t lprt;
+	int                sock  = -1; /* mark socket invalid */
+	int                retry = MAX_RETRIES;
+	u_int16_t          lprt  = lrng;
 
 	if (phls == NULL || ctyp == NULL)	/* Sanity check	*/
 		misc_die(FL, "socket_d_connect: ?phls? ?ctyp?");
 
-	/*
-	** First of all, get a socket
-	*/
-	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
-		syslog_error("can't create %s socket", ctyp);
-		exit(EXIT_FAILURE);
-	}
-	socket_opts(sock, SK_DATA);
+	while(0 <= retry--)
+	{
+		/*
+		** First of all, get a socket
+		*/
+		if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
+			syslog_error("can't create %s socket", ctyp);
+			exit(EXIT_FAILURE);
+		}
+		socket_opts(sock, SK_DATA);
 
-	/*
-	** Bind the socket, taking care of a given port range
-	*/
-	if (lrng > 0 && urng > 0) {
-		memset(&saddr, 0, sizeof(saddr));
-		saddr.sin_addr.s_addr = htonl(ladr);
-		saddr.sin_family      = AF_INET;
-
-		for (lprt = lrng; lprt <= urng; lprt++) {
+		/*
+		** check if we have to use a port range
+		*/
+		if( !(INPORT_ANY == lrng && INPORT_ANY == urng)) {
+			/*
+			** Bind the socket, taking care of a given port range
+			*/
+			if(incr) {
 #if defined(COMPILE_DEBUG)
-			debug(2, "try to con-bind %s to %s:%d", ctyp,
-				socket_addr2str(ladr), (int) lprt);
+				debug(2, "%s: about to bind to %s:range(%d-%d)",
+				         ctyp, socket_addr2str(ladr),
+				         lprt, urng);
+#endif
+				lprt = socket_d_bind(sock, ladr,
+				                     lprt, urng, incr);
+			} else {
+#if defined(COMPILE_DEBUG)
+				debug(2, "%s: about to bind to %s:range(%d-%d)",
+				         ctyp, socket_addr2str(ladr),
+				         lrng, urng);
+#endif
+				lprt = socket_d_bind(sock, ladr,
+				                     lrng, urng, incr);
+			}
+			if (INPORT_ANY == lprt) {
+				/* nothing found? */
+				close(sock);
+				return 0;
+			}
+		} else lprt = INPORT_ANY;
+
+		/*
+		** Actually connect the socket
+		*/
+		memset(&saddr, 0, sizeof(saddr));
+		saddr.sin_addr.s_addr = htonl(addr);
+		saddr.sin_family      = AF_INET;
+		saddr.sin_port        = htons(port);
+
+		if (connect(sock, (struct sockaddr *) &saddr,
+		                  sizeof(saddr)) < 0)
+		{
+#if defined(COMPILE_DEBUG)
+			debug(2, "%s: connect failed with '%s'",
+			         ctyp, strerror(errno));
 #endif
-			saddr.sin_port = htons(lprt);
-			if (bind(sock, (struct sockaddr *) &saddr,
-						sizeof(saddr)) == 0)
-				break;
-			if (errno != EADDRINUSE)
-				lprt = urng + 1;
-		}
-		if (lprt > urng) {	/* nothing found? */
 			close(sock);
-			return 0;
-		}
+			sock = -1;
+			/* check if is makes sense to retry?
+			** perhaps we only need an other
+			** local port (EADDRNOTAVAIL) for
+			** this destination?
+			*/
+			if( !(EINTR == errno ||
+			      EAGAIN == errno ||
+			      EADDRINUSE == errno ||
+			      EADDRNOTAVAIL == errno))
+			{
+				/*
+				** an other (real) error ocurred
+				*/
+				return 0;
+			} else
+			if(incr && INPORT_ANY != lprt) {
+				/*
+				** increment lower range if we use
+				** increment mode and have a range
+				*/
+				if(lprt < urng) {
+					lprt++; /* incr lower range */
+				} else {
+				/*
+				** no more ports in range we can try
+				*/
+					return 0;
+				}
+			}
+		} else break;
 	}
 
 	/*
-	** Actually connect the socket
+	** check if we have a valid, connected socket
 	*/
-	memset(&saddr, 0, sizeof(saddr));
-	saddr.sin_addr.s_addr = htonl(addr);
-	saddr.sin_family      = AF_INET;
-	saddr.sin_port        = htons(port);
-
-	if (connect(sock, (struct sockaddr *) &saddr,
-				sizeof(saddr)) < 0) {
+	if(-1 == sock) {
 		close(sock);
 		return 0;
 	}
@@ -1407,7 +1743,8 @@ u_int16_t socket_str2port(char *name, u_int16_t dflt)
 **
 **	Return........:	Dotted decimal string for addr
 **
-**	Purpose.......: Convert IP address to human readable form.
+**	Purpose.......: Convert IP address (host byte order) into
+**	                human readable form.
 **			The buffer is reused in subsequent calls,
 **			so the caller better move the result away.
 **
@@ -1420,7 +1757,7 @@ char *socket_addr2str(u_int32_t addr)
 
 	memset(&iadr, 0, sizeof(iadr));
 	iadr.s_addr = htonl(addr);
-	strcpy(str, inet_ntoa(iadr));
+	misc_strncpy(str, inet_ntoa(iadr), sizeof(str));
 	return str;
 }
 
@@ -1433,10 +1770,12 @@ char *socket_addr2str(u_int32_t addr)
 **			peer		LOC_END or REM_END
 **			port		Pointer to port
 **
-**	Return........:	IP address in host order
+**	Return........:	IP address in host byte order
+**			or INADDR_NONE on failure
 **
 **	Purpose.......: Retrieve the IP address of a socket,
-**			for either the peer or the local end.
+**			for either the peer or the local end
+**			in host byte order.
 **
 ** ------------------------------------------------------------ */
 
@@ -1460,14 +1799,14 @@ u_int32_t socket_sck2addr(int sock, int peer, u_int16_t *port)
 	}
 	if (r < 0) {
 		syslog_error("can't get %sname for socket %d", s, sock);
-		exit(EXIT_FAILURE);
+		return INADDR_NONE;
 	}
 
 	/*
 	** Return the port if requested
 	*/
 	if (port != NULL)
-		*port = (u_int16_t) htons(saddr.sin_port);
+		*port = (u_int16_t) ntohs(saddr.sin_port);
 
 	/*
 	** Return the address part
@@ -1476,8 +1815,338 @@ u_int32_t socket_sck2addr(int sock, int peer, u_int16_t *port)
 }
 
 
+/* ------------------------------------------------------------ **
+**
+**	Function......:	socket_chkladdr
+**
+**	Parameters....:	addr		ip address to check
+**
+**	Return........:	0 if not found, 1 if found, -1 on error
+**
+**	Purpose.......: Check if addr (in network byte order)
+**			is used on an local network interface.
+**
+** ------------------------------------------------------------ */
+
+int        socket_chkladdr(u_int32_t addr)
+{
+#define DEFAULT_IFNUM	512
+	struct ifconf  ifc;
+	int            ifn = DEFAULT_IFNUM;
+	int            i, sock;
+
+	sock = socket(AF_INET, SOCK_DGRAM, 0);
+	if(-1 == sock) {
+#if defined(COMPILE_DEBUG)
+		debug(2, "can not create socket: %s", NIL(strerror(errno)));
+#endif
+		return -1;
+	}
+
+#if defined(SIOCGIFNUM)
+	if( ioctl(sock, SIOCGIFNUM, (char *) &ifn) < 0) {
+#if defined(COMPILE_DEBUG)
+		debug(2, "ioctl SIOCGIFNUM failed: %s", NIL(strerror(errno)));
+#endif
+		ifn = DEFAULT_IFNUM;	/* ignore failure */
+	}
+#endif	/* SIOCGIFNUM */
+
+	ifc.ifc_len = ifn * sizeof (struct ifreq);
+	ifc.ifc_buf = malloc(ifc.ifc_len);
+	if( !ifc.ifc_buf) {
+#if defined(COMPILE_DEBUG)
+		debug(2, "malloc(ifc.ifc_len=%d) failed: %s",
+			 ifc.ifc_len, NIL(strerror(errno)));
+#endif
+		close(sock);
+		return -1;
+	}
+        memset(ifc.ifc_buf, 0, ifc.ifc_len);
+
+	if( ioctl(sock, SIOCGIFCONF, (char *)&ifc) < 0) {
+#if defined(COMPILE_DEBUG)
+		debug(2, "ioctl SIOCGIFCONF failed: %s", NIL(strerror(errno)));
+#endif
+		free(ifc.ifc_buf);
+		close(sock);
+		return -1;
+	}
+	close(sock);
+
+	for( i=0; i<ifc.ifc_len; ) {
+		struct ifreq       *ifr = (struct ifreq *) &ifc.ifc_buf[i];
+		struct sockaddr_in *sa  = (struct sockaddr_in*) &ifr->ifr_addr;
+
+		i += sizeof( *ifr);
+
+		if(AF_INET     != ifr->ifr_addr.sa_family
+		|| INADDR_ANY  == sa->sin_addr.s_addr
+		|| INADDR_NONE == sa->sin_addr.s_addr)
+			continue;
+
+#if defined(COMPILE_DEBUG)
+		debug(4, "interface %s has ip-address: %s",
+			 ifr->ifr_name, inet_ntoa(sa->sin_addr));
+#endif
+
+		if( sa->sin_addr.s_addr == addr) {
+#if defined(COMPILE_DEBUG)
+			debug(2, "found local ip addr: %s",
+				 inet_ntoa(sa->sin_addr));
+#endif
+			free(ifc.ifc_buf);
+			return 1;
+		}
+	}
+	free(ifc.ifc_buf);
+
+#if defined(COMPILE_DEBUG)
+	debug(2, "requested ip addr is not a local one");
+#endif
+	return 0;
+}
+
+
+/* ------------------------------------------------------------ **
+**
+**	Function......:	socket_orgdst
+**
+**	Parameters....:	phls	pointer to high-level socket
+**			addr	pointer to in_addr_t
+**			port	pointer to in_port_t
+**
+**	Return........:	0 on success, -1 on error
+**
+**	Purpose.......: get the original (transparent) destination
+**	                addr and port the peer / client wanted to
+**	                connect (network byte order!).
+**
+** ------------------------------------------------------------ */
+
+int socket_orgdst(HLS *phls, u_int32_t *addr, u_int16_t *port)
+{
+	struct sockaddr_in name;
+#if defined(HAVE_LINUX_NETFILTER_IPV4_H) && defined(SO_ORIGINAL_DST)
+	struct sockaddr_in dest;
+#endif
+#if defined(HAVE_NETINET_IP_NAT_H) && defined(SIOCGNATL)
+	natlookup_t        natlook, *nlptr = &natlook;
+	int                rc, nat_fd = -1;
+#endif
+	socklen_t          len;
+
+	/*
+	** sanity args checks
+	*/
+	if( !(phls && -1 != phls->sock && addr && port))
+		return -1;
+
+	len = sizeof(name);
+	memset(&name, 0, len);
+	if(getsockname(phls->sock, (struct sockaddr *)&name, &len) < 0) {
+		syslog_error("can't get sockname for socket %d", phls->sock);
+		return -1;
+	}
+
+#if defined(HAVE_LINUX_NETFILTER_IPV4_H) && defined(SO_ORIGINAL_DST)
+	/*
+	** IP-Tables uses SO_ORIGINAL_DST getsockopt call
+	*/
+	len = sizeof(dest);
+	memset(&dest, 0, len);
+	if(getsockopt(phls->sock, SOL_IP, SO_ORIGINAL_DST, &dest, &len) < 0) {
+		/*
+		** ENOPROTOOPT: 2.2 kernel ... no iptables support
+		** ENOENT     : 2.4 kernel without iptables support
+		*/
+		if(ENOPROTOOPT == errno || ENOENT == errno) {
+			if(ENOENT == errno)
+			syslog_write(T_DBG,
+			"iptables not supported or ipchains support active");
+		} else {
+			syslog_error(
+			"can't get iptables transparent proxy destination");
+			return -1;
+		}
+	} else {
+#if defined(COMPILE_DEBUG)
+		debug(2, "iptables transparent proxy destination: %s:%d",
+			 socket_addr2str(ntohl(dest.sin_addr.s_addr)),
+			 ntohs(dest.sin_port));
+#endif
+		if((name.sin_port == dest.sin_port) &&
+		   (name.sin_addr.s_addr == dest.sin_addr.s_addr)) {
+			syslog_write(T_DBG,
+			       "iptables proxy destination %s:%d is local",
+			       socket_addr2str(ntohl(dest.sin_addr.s_addr)),
+			       ntohs(dest.sin_port));
+			return -1;
+		}
+		*addr = dest.sin_addr.s_addr;
+		*port = dest.sin_port;
+		return 0;
+	}
+#endif
+
+#if defined(HAVE_NETINET_IP_NAT_H) && defined(SIOCGNATL)
+
+	/*
+	** BSD ipnat table lookup
+	*/
+	if ((nat_fd = open(IPL_NAT, O_RDONLY, 0)) < 0) {
+		syslog_error("can't open ipnat device '%.*s'",
+		             MAX_PATH_SIZE, IPL_NAT);
+		return -1;
+	}
+
+	memset(&natlook, 0, sizeof(natlook));
+	natlook.nl_flags        = IPN_TCP;
+	natlook.nl_inip.s_addr  = name.sin_addr.s_addr;
+	natlook.nl_inport       = name.sin_port;
+	natlook.nl_outip.s_addr = htonl(phls->addr);
+	natlook.nl_outport      = htons(phls->port);
+
+	/*
+	** handle versions differences...
+	*/
+	rc = 0;
+	if(63 == (SIOCGNATL & 0xff)) {
+		rc = ioctl(nat_fd, SIOCGNATL, &nlptr);
+	} else {
+		rc = ioctl(nat_fd, SIOCGNATL, &natlook);
+	}
+	close(nat_fd);
+
+	if(rc < 0) {
+		if(errno != ESRCH) {
+			syslog_error("can't get ipnat transparent "
+			             "proxy destination");
+		}
+		return -1;
+	}
+
+#if defined(COMPILE_DEBUG)
+	debug(2, "ipnat transparent proxy destination: %s:%d",
+		       socket_addr2str(ntohl(natlook.nl_realip.s_addr)),
+		       ntohs(natlook.nl_realport));
+#endif
+
+	if((natlook.nl_realport == name.sin_port) &&
+	   (natlook.nl_realip.s_addr == name.sin_addr.s_addr))
+	{
+		syslog_write(T_DBG,
+		       "ipnat proxy destination %s:%d is local",
+		       socket_addr2str(ntohl(natlook.nl_realip.s_addr)),
+		       ntohs(natlook.nl_realport));
+		return -1;
+	} else {
+		*addr = natlook.nl_realip.s_addr;
+		*port = natlook.nl_realport;
+		return 0;
+	}
+
+#else /* !BSD-IPNAT */
+
+	/*
+	** IP-Chains uses getsockname, as "transparent address"
+	*/
+
+#if defined(COMPILE_DEBUG)
+		debug(2, "ipchains transparent proxy destination: %s:%d",
+			 socket_addr2str(ntohl(name.sin_addr.s_addr)),
+			 ntohs(name.sin_port));
+#endif
+
+	*addr = name.sin_addr.s_addr;
+	*port = name.sin_port;
+
+#endif
+	return 0;
+}
+
+
+/* ------------------------------------------------------------ **
+**
+**	Function......:	getfqhostname
+**
+**	Parameters....:	fqhost	buffer to store the host name
+**			n	size of the buffer
+**
+**	Return........:	0 on success, -1 on error
+**
+**	Purpose.......: get the full qualified (resolved)
+**                      host name of the current/local host
+**
+** ------------------------------------------------------------ */
+
+int getfqhostname(char *fqhost, size_t n)
+{
+	char           hname[MAXHOSTNAMELEN];
+	struct hostent *hp;
+
+	if( !(n > 0 && fqhost))
+		return -1;
+
+	memset(hname, 0, sizeof(hname));
+	if( gethostname(hname, sizeof(hname)-1))
+		return -1;
+
+	if( !(hp = gethostbyname(hname)))
+		return -1;
+
+	misc_strncpy(fqhost, hp->h_name, n);
+
+	return 0;
+}
+
+/* ------------------------------------------------------------ **
+**
+**	Function......:	getfqdomainname
+**
+**	Parameters....:	fqhost	buffer to store the domain name
+**			n	size of the buffer
+**
+**	Return........:	0 on success, -1 on error
+**
+**	Purpose.......: get the full qualified (resolved)
+**                      domain name of the current/local host
+**
+** ------------------------------------------------------------ */
+
+int getfqdomainname(char *fqdomain, size_t n)
+{
+	char hname[MAXHOSTNAMELEN], *p;
+
+	if( !(n > 0 && fqdomain))
+		return -1;
+
+	if(getfqhostname(hname, sizeof(hname)))
+		return -1;
+
+	p = strchr(hname, (int)'.');
+	if(p && *(p+1)) {
+		misc_strncpy(fqdomain, p+1, n);
+		return 0;
+	}
+
+	return -1;
+}
+
 /* ------------------------------------------------------------
  * $Log: com-socket.c,v $
+ * Revision 1.6  2002/01/14 18:26:55  mt
+ * implemented socket_orgdst to read transparent proxying destinations
+ * implemented a MaxRecvBufSize option limiting max recv buffer size
+ * implemented workarround for Netscape (4.x) directory symlink handling
+ * extended log messages to provide basic transfer statistics data
+ * fixed socket_gets to wait for a complete line if no EOL found
+ * added snprintf usage, replaced strcpy/strncpy with misc_strncpy
+ *
+ * Revision 1.5  2001/11/06 23:04:43  mt
+ * applied / merged with transparent proxy patches v8
+ * see ftp-proxy/NEWS for more detailed release news
+ *
  * Revision 1.4  1999/09/21 05:42:28  wiegand
  * syslog / abort review
  *
