@@ -1,5 +1,5 @@
 /*
- * $Id: ftp-client.c,v 1.6 1999/10/19 11:04:51 wiegand Exp $
+ * $Id: ftp-client.c,v 1.8 2002/01/14 19:35:44 mt Exp $
  *
  * Functions for the FTP Proxy client handling
  *
@@ -30,7 +30,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: ftp-client.c,v 1.6 1999/10/19 11:04:51 wiegand Exp $";
+static char rcsid[] = "$Id: ftp-client.c,v 1.8 2002/01/14 19:35:44 mt Exp $";
 #endif
 
 #include <config.h>
@@ -111,6 +111,9 @@ static RETSIGTYPE client_signal(int signo)
 	close_flag  = 1;
 
 	signal(signo, client_signal);
+#if RETSIGTYPE != void
+	return 0;
+#endif
 }
 
 
@@ -128,10 +131,11 @@ static RETSIGTYPE client_signal(int signo)
 
 void client_run(void)
 {
-	int sock, need, diff;
-	char str[1024], *p, *q;
+	int  sock, need, diff, rw_prio;
+	char str[MAX_PATH_SIZE * 2];
+	char *p, *q;
 	FILE *fp;
-	BUF *buf;
+	BUF  *buf;
 
 	/*
 	** Setup client signal handling (mostly graceful exit)
@@ -147,6 +151,7 @@ void client_run(void)
 	** Prepare our general client context
 	*/
 	memset(&ctx, 0, sizeof(ctx));
+	ctx.sess_beg = time(NULL);
 	ctx.cli_mode = MOD_ACT_FTP;
 	ctx.expect   = EXP_IDLE;
 	ctx.timeout  = config_int(NULL, "TimeOut", 900);
@@ -200,21 +205,15 @@ void client_run(void)
 	*/
 	if ((p = config_str(NULL, "WelcomeString", NULL)) == NULL)
 		p = "%h FTP server (Version %v - %b) ready";
-	strcpy(str, socket_msgline(p));
+	misc_strncpy(str, socket_msgline(p), sizeof(str));
 	client_respond(220,
 		config_str(NULL, "WelcomeMessage", NULL), str);
 
 	/*
-	** See if we need to change our root directory
+	** Set the "read only if nothing to write there" flag
+	** for data transfers if a read bufffer limit is set.
 	*/
-	chdir("/");
-	if ((p = config_str(NULL, "ServerRoot", NULL)) != NULL) {
-		if (chroot(p)) {
-			syslog_error("can't chroot to '%.1024s'", p);
-			exit(EXIT_FAILURE);
-		}
-		chdir("/");
-	}
+	rw_prio = config_int(NULL, "MaxRecvBufSize", 0);
 
 	/*
 	** Enter the client mainloop
@@ -222,14 +221,38 @@ void client_run(void)
 	while (close_flag == 0) {
 		/*
 		** We need to go into select() only
-		** if all input has been processed.
-		** (Data buffers are never split.)
+		** if all input has been processed
+		**   or
+		** we wait for more data to get a line
+		** complete (partially sent, no EOL).
+		**
+		** (data buffers are never splited)
 		*/
 		need = 1;
 		if (ctx.cli_ctrl && ctx.cli_ctrl->rbuf)
 			need = 0;
 		if (ctx.srv_ctrl && ctx.srv_ctrl->rbuf)
 			need = 0;
+		if((ctx.cli_ctrl && ctx.cli_ctrl->more>0) ||
+		   (ctx.srv_ctrl && ctx.srv_ctrl->more>0))
+			need = 1;
+
+		/*
+		** use higher priority to writes
+		*/
+		if(rw_prio > 0 && ctx.srv_data && ctx.cli_data) {
+			if(ctx.srv_data->wbuf) {
+				ctx.cli_data->more = -1;
+			} else {
+				ctx.cli_data->more = 0;
+			}
+			if(ctx.cli_data->wbuf) {
+				ctx.srv_data->more = -1;
+			} else {
+				ctx.srv_data->more = 0;
+			}
+		}
+
 		if (need != 0) {
 			if (socket_exec(ctx.timeout, &close_flag) <= 0)
 				break;	/* Timed out or worse */
@@ -300,11 +323,30 @@ void client_run(void)
 			diff = (int) (time(NULL) - ctx.xfer_beg);
 			if (diff < 1)
 				diff = 1;
+
 			syslog_write(U_INF,
-				"Transfer for %s: %s '%s' %d sec",
+				"Transfer for %s: %s '%s' %s %u/%d byte/sec",
 				ctx.cli_ctrl->peer, ctx.xfer_cmd,
-				ctx.xfer_arg, diff);
-			client_data_reset();
+				ctx.xfer_arg,
+				ctx.cli_data->rcnt ? "sent" : "read",
+				ctx.cli_data->rcnt ? ctx.cli_data->rcnt
+				                   : ctx.cli_data->wcnt,
+				diff);
+
+			/*
+			** update statistics data
+			*/
+			if(ctx.cli_data->rcnt)
+				ctx.xfer_rsec += diff;
+			ctx.xfer_rcnt += ctx.cli_data->rcnt;
+			if(ctx.cli_data->wcnt)
+				ctx.xfer_wsec += diff;
+			ctx.xfer_wcnt += ctx.cli_data->wcnt;
+
+			/*
+			** data transfer reset
+			*/
+			client_data_reset(MOD_RESET);
 
 			/*
 			** Doom the corresponding server socket
@@ -320,14 +362,21 @@ void client_run(void)
 		}
 
 		if (ctx.srv_data != NULL && ctx.srv_data->sock == -1) {
+
 #if defined(COMPILE_DEBUG)
 			debug(3, "about to destroy Srv-Data");
 #endif
 			/*
-			** Doom the corresponding client socket
+			** Doom the corresponding client socket _except_
+			** FailResetsPasv=no and we expect PASV response
 			*/
-			if (ctx.cli_data != NULL)
-				ctx.cli_data->kill = 1;
+			if(ctx.cli_data != NULL) {
+				if(config_bool(NULL,"FailResetsPasv", 0)) {
+					ctx.cli_data->kill = 1;
+				} else if(ctx.expect != EXP_PASV) {
+					ctx.cli_data->kill = 1;
+				}
+			}
 
 			/*
 			** Don't forget to remove the dead socket
@@ -391,6 +440,19 @@ void client_run(void)
 		/* at this point the main loop resumes ... */
 	}
 
+	/*
+	** Display basic session statistics...
+	**   in secs since session begin
+	**   downloads / read (xfer-reads from server)
+	**   uploads   / send (xfer-sends from server)
+	*/
+	syslog_write(U_INF, "closing connect from %s after %d secs - "
+	                    "read %d/%d, sent %d/%d byte/sec",
+	             ctx.cli_ctrl ? ctx.cli_ctrl->peer : "unknown peer",
+	             time(NULL)-ctx.sess_beg,
+	             ctx.xfer_wcnt, ctx.xfer_wsec,
+	             ctx.xfer_rcnt, ctx.xfer_rsec);
+
 	/*
 	** Free allocated memory
 	*/
@@ -445,7 +507,10 @@ static void client_cli_ctrl_read(char *str)
 					ctx.cli_ctrl->peer);
 				socket_printf(ctx.cli_ctrl,
 					"%c%c%c", IAC, DONT, arg[2]);
-				memmove(arg, arg + 3, strlen(arg) - 2);
+				if(arg[2])
+					memmove(arg, arg + 3, strlen(arg) - 2);
+				else
+					memmove(arg, arg + 1, strlen(arg));
 				break;
 
 			case DO:
@@ -458,7 +523,10 @@ static void client_cli_ctrl_read(char *str)
 					ctx.cli_ctrl->peer);
 				socket_printf(ctx.cli_ctrl,
 					"%c%c%c", IAC, WONT, arg[2]);
-				memmove(arg, arg + 3, strlen(arg) - 2);
+				if(arg[2])
+					memmove(arg, arg + 3, strlen(arg) - 2);
+				else
+					memmove(arg, arg + 1, strlen(arg));
 				break;
 
 			case IAC:
@@ -470,11 +538,11 @@ static void client_cli_ctrl_read(char *str)
 				syslog_write(U_INF, "IAC-%s from %s",
 						(c == IP) ? "IP" : "DM",
 						ctx.cli_ctrl->peer);
-				memmove(str, arg + 2, strlen(arg) + 1);
+				memmove(arg, arg + 2, strlen(arg) - 1);
 				break;
 
 			default:
-				memmove(arg, arg + 2, strlen(arg) + 1);
+				memmove(arg, arg + 1, strlen(arg));
 		}
 	}
 
@@ -515,7 +583,7 @@ static void client_cli_ctrl_read(char *str)
 			cmd->legal = 1;		/* Need this one! */
 		if (strcasecmp(str, cmd->name) != 0)
 			continue;
-		if (cmd->legal == 0) {
+		if ((cmd->legal == 0) && strcasecmp("QUIT", cmd->name)) {
 			client_respond(502, NULL, "'%.32s': "
 				"command not implemented", str);
 			syslog_write(U_WRN,
@@ -640,29 +708,37 @@ static void client_srv_ctrl_read(char *str)
 			** 230=logged in, 331=need password
 			*/
 			socket_printf(ctx.cli_ctrl, "%s\r\n", str);
-			if (c1 != 2 && c1 != 3)
+			if (c1 != 2 && c1 != 3) {
 				ctx.cli_ctrl->kill = 1;
+			}
 			ctx.expect = EXP_IDLE;
 			break;
 
+		case EXP_ABOR:
+			if (c1 == 2) {
+				client_data_reset(MOD_RESET);
+				ctx.expect = EXP_IDLE;
+			}
+			break;
+
 		case EXP_PASV:
-			if (code == 227 && *arg != '\0')
+			if (code == 227 && *arg != '\0') {
 				client_srv_passive(arg);
-			else {
+			} else {
 				socket_printf(ctx.cli_ctrl,
 						"%s\r\n", str);
-				client_data_reset();
+				client_data_reset(MOD_RESET);
 				ctx.expect = EXP_IDLE;
 			}
 			break;
 
 		case EXP_PORT:
-			if (code == 200)
+			if (code == 200) {
 				client_xfer_fireup();
-			else {
+			} else {
 				socket_printf(ctx.cli_ctrl,
 						"%s\r\n", str);
-				client_data_reset();
+				client_data_reset(MOD_RESET);
 				ctx.expect = EXP_IDLE;
 			}
 			break;
@@ -677,18 +753,15 @@ static void client_srv_ctrl_read(char *str)
 			} else {
 				socket_printf(ctx.cli_ctrl,
 						"%s\r\n", str);
-				client_data_reset();
+				if(config_bool(NULL,"FailResetsPasv", 0)) {
+					client_data_reset(MOD_RESET);
+				} else {
+					client_data_reset(ctx.cli_mode);
+				}
 			}
 			ctx.expect = EXP_IDLE;
 			break;
 
-		case EXP_ABOR:
-			if (c1 == 2) {
-				client_data_reset();
-				ctx.expect = EXP_IDLE;
-			}
-			break;
-
 		case EXP_PTHR:
 			socket_printf(ctx.cli_ctrl, "%s\r\n", str);
 			ctx.expect = EXP_IDLE;
@@ -701,7 +774,7 @@ static void client_srv_ctrl_read(char *str)
 					"server closed connection "
 					"for %s", ctx.cli_ctrl->peer);
 				ctx.cli_ctrl->kill = 1;
-			}else {
+			} else {
 				syslog_write(T_WRN,
 					"bogus '%.512s' from "
 					"Server-PI for %s",
@@ -729,6 +802,7 @@ static void client_srv_passive(char *arg)
 	int h1, h2, h3, h4, p1, p2;
 	u_int32_t addr, ladr;
 	u_int16_t port;
+	int       incr;
 
 	if (arg == NULL)		/* Basic sanity check	*/
 		return;
@@ -741,26 +815,33 @@ static void client_srv_passive(char *arg)
 		arg++;
 	if (sscanf(arg, "%d,%d,%d,%d,%d,%d",
 			&h1, &h2, &h3, &h4, &p1, &p2) != 6) {
-		syslog_error("bad 227 from server for %s",
+		syslog_error("bad PASV 277 response from server for %s",
 						ctx.cli_ctrl->peer);
 		client_respond(425, NULL, "Can't open data connection");
-		client_data_reset();
+		client_data_reset(MOD_RESET);
 		ctx.expect = EXP_IDLE;
 		return;
 	}
 	addr = (u_int32_t) ((h1 << 24) + (h2 << 16) + (h3 << 8) + h4);
 	port = (u_int16_t) ((p1 <<  8) +  p2);
 
+	/*
+	** should we bind a rand(port-range) or increment?
+	*/
+	incr = !config_bool(NULL,"SockBindRand", 0);
+
 	/*
 	** Open a connection to the server at the given port
 	*/
 	ladr = socket_sck2addr(ctx.srv_ctrl->sock, LOC_END, NULL);
 	if (socket_d_connect(addr, port, ladr, ctx.srv_lrng,
-			ctx.srv_urng, &(ctx.srv_data), "Srv-Data") == 0) {
+			ctx.srv_urng, &(ctx.srv_data),
+			"Srv-Data", incr) == 0)
+	{
 		syslog_error("can't connect Srv-Data for %s",
 						ctx.cli_ctrl->peer);
 		client_respond(425, NULL, "Can't open data connection");
-		client_data_reset();
+		client_data_reset(MOD_RESET);
 		ctx.expect = EXP_IDLE;
 		return;
 	}
@@ -789,22 +870,40 @@ static void client_srv_passive(char *arg)
 
 static void client_xfer_fireup(void)
 {
-	u_int32_t ladr;
+	u_int32_t ladr = INADDR_ANY;
+	int       incr;
+
+	/*
+	** should we bind a rand(port-range) or increment?
+	*/
+	incr = !config_bool(NULL,"SockBindRand", 0);
 
 	/*
 	** If appropriate, connect to the client's data port
 	*/
 	if (ctx.cli_mode == MOD_ACT_FTP) {
-		ladr = socket_sck2addr(ctx.cli_ctrl->sock,
+		/*
+		** TransProxy mode: check if we can use our real
+		** ip instead of the server's one as our local ip,
+		** we pre-bind the socket/ports to before connect.
+		*/
+		if(config_bool(NULL, "AllowTransProxy", 0)) {
+			ladr = config_addr(NULL, "Listen",
+					(u_int32_t)INADDR_ANY);
+		}
+		if(INADDR_ANY == ladr) {
+			ladr = socket_sck2addr(ctx.cli_ctrl->sock,
 						LOC_END, NULL);
+		}
 		if (socket_d_connect(ctx.cli_addr, ctx.cli_port,
 				ladr, ctx.act_lrng, ctx.act_urng,
-				&(ctx.cli_data), "Cli-Data") == 0) {
+				&(ctx.cli_data), "Cli-Data", incr) == 0)
+		{
 			syslog_error("can't connect Cli-Data for %s",
 						ctx.cli_ctrl->peer);
 			client_respond(425, NULL,
 					"Can't open data connection");
-			client_data_reset();
+			client_data_reset(MOD_RESET);
 			ctx.expect = EXP_IDLE;
 			return;
 		}
@@ -829,7 +928,6 @@ static void client_xfer_fireup(void)
 	*/
 	memset(ctx.xfer_rep, 0, sizeof(ctx.xfer_rep));
 	ctx.xfer_beg = time(NULL);
-	ctx.xfer_cnt = 0;
 
 	ctx.expect = EXP_XFER;		/* Expect 226 complete	*/
 }
@@ -852,7 +950,7 @@ static void client_xfer_fireup(void)
 void client_respond(int code, char *file, char *fmt, ...)
 {
 	va_list aptr;
-	char str[1024], *p, *q;
+	char str[MAX_PATH_SIZE * 2], *p, *q;
 	FILE *fp;
 
 	/*
@@ -873,7 +971,11 @@ void client_respond(int code, char *file, char *fmt, ...)
 	** The last line carries the ultimate reponse code
 	*/
 	va_start(aptr, fmt);
+#if defined(HAVE_VSNPRINTF)
+	vsnprintf(str, sizeof(str), fmt, aptr);
+#else
 	vsprintf(str, fmt, aptr);
+#endif
 	va_end(aptr);
 	socket_printf(ctx.cli_ctrl, "%03d %s.\r\n", code, str);
 }
@@ -908,7 +1010,7 @@ void client_reinit(void)
 		socket_kill(ctx.srv_ctrl);
 		ctx.srv_ctrl = NULL;
 	}
-	client_data_reset();
+	client_data_reset(MOD_RESET);
 
 	/*
 	** Remove the current user and status
@@ -925,7 +1027,8 @@ void client_reinit(void)
 **
 **	Function......:	client_data_reset
 **
-**	Parameters....:	(none)
+**	Parameters....:	mode	reset client transfer mode
+**				default value is MOD_RESET
 **
 **	Return........:	(none)
 **
@@ -933,14 +1036,19 @@ void client_reinit(void)
 **
 ** ------------------------------------------------------------ */
 
-void client_data_reset(void)
+void client_data_reset(int mode)
 {
 	memset(ctx.xfer_cmd, 0, sizeof(ctx.xfer_cmd));
 	memset(ctx.xfer_arg, 0, sizeof(ctx.xfer_arg));
 	ctx.xfer_beg = 0;
-	ctx.xfer_cnt = 0;
 
-	ctx.cli_mode = MOD_ACT_FTP;
+	/*
+	** reset client transfer mode to the specified one
+	** or (mode==MOD_RESET) to default mode MOD_ACT_FTP
+	**
+	** Note: a reset to default is the normal behaviour
+	*/
+	ctx.cli_mode = mode ? mode : MOD_ACT_FTP;
 
 	if (ctx.cli_ctrl != NULL) {
 		ctx.cli_addr = ctx.cli_ctrl->addr;
@@ -951,6 +1059,16 @@ void client_data_reset(void)
 
 /* ------------------------------------------------------------
  * $Log: ftp-client.c,v $
+ * Revision 1.8  2002/01/14 19:35:44  mt
+ * implemented workarround for Netscape (4.x) directory symlink handling
+ * implemented a MaxRecvBufSize option limiting max recv buffer size
+ * extended log messages to provide basic transfer statistics data
+ * added snprintf usage if supported, replaced strncpy with misc_strncpy
+ *
+ * Revision 1.7  2001/11/06 23:04:43  mt
+ * applied / merged with transparent proxy patches v8
+ * see ftp-proxy/NEWS for more detailed release news
+ *
  * Revision 1.6  1999/10/19 11:04:51  wiegand
  * make sure transfer time has reasonable value
  *
