@@ -1,5 +1,5 @@
 /*
- * $Id: ftp-cmds.c,v 1.7 1999/10/19 10:19:15 wiegand Exp $
+ * $Id: ftp-cmds.c,v 1.9 2002/01/14 19:39:30 mt Exp $
  *
  * FTP Proxy command handling
  *
@@ -30,7 +30,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: ftp-cmds.c,v 1.7 1999/10/19 10:19:15 wiegand Exp $";
+static char rcsid[] = "$Id: ftp-cmds.c,v 1.9 2002/01/14 19:39:30 mt Exp $";
 #endif
 
 #include <config.h>
@@ -90,6 +90,9 @@ static void cmds_port(CONTEXT *ctx, char *arg);
 static void cmds_pasv(CONTEXT *ctx, char *arg);
 static void cmds_xfer(CONTEXT *ctx, char *arg);
 static void cmds_abor(CONTEXT *ctx, char *arg);
+#if defined(ENABLE_SSL) /* <!-- SSL --> */
+static void cmds_auth(CONTEXT *ctx, char *arg);
+#endif /* <!-- /SSL --> */
 #if defined(ENABLE_RFC1579)
 static void cmds_apsv(CONTEXT *ctx, char *arg);
 #endif
@@ -155,6 +158,9 @@ static CMD cmdlist[] = {
 	{ "XRMD", cmds_pthr, REST },
 	{ "XPWD", cmds_pthr, REST },
 	{ "XCUP", cmds_pthr, REST },
+#if defined(ENABLE_SSL) /* <!-- SSL --> */
+	{ "AUTH", cmds_auth, REST },	/* Only needed for SSL	*/
+#endif /* <!-- /SSL --> */
 #if defined(ENABLE_RFC1579)
 	{ "APSV", cmds_apsv, REST },	/* As per RFC 1579	*/
 #endif
@@ -247,11 +253,11 @@ void cmds_set_allow(char *allow)
 	** Scan the allow list and enable accordingly
 	*/
 	for (p = allow; *p != '\0'; ) {
-		while (*p != '\0' && isalpha(*p) == 0)
+		while (*p != '\0' && isalpha((int)*p) == 0)
 			p++;
 		if (*p == '\0')
 			break;
-		for (q = p, i = 0; isalpha(*q); q++, i++)
+		for (q = p, i = 0; isalpha((int)*q); q++, i++)
 			;
 		for (cmd = cmdlist; cmd->name; cmd++) {
 			if (cmd->len != i)
@@ -340,11 +346,12 @@ static void cmds_pthr(CONTEXT *ctx, char *arg)
 
 static void cmds_user(CONTEXT *ctx, char *arg)
 {
-	char *p, *q;
-	CMD *cmd;
-	int sock;
 	struct sockaddr_in saddr;
-	u_int16_t lprt;
+	int                sock, len;
+	int                incr, retry;
+	u_int16_t          lprt, lowrng, bind_res;
+	char               *p, *q;
+	CMD                *cmd;
 
 	if (ctx == NULL)		/* Basic sanity check	*/
 		misc_die(FL, "cmds_user: ?ctx?");
@@ -364,6 +371,7 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 	*/
 	client_reinit();
 
+#if defined(HAVE_REGEX)
 	/*
 	** Check for a RegEx constraint on the USER command
 	*/
@@ -385,15 +393,62 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 		}
 		break;
 	}
+#endif
+
+	/*
+	** Check for permission and existence of "transparent proxy"
+	** magic destination address and port from the client socket
+	*/
+	ctx->magic_addr = INADDR_ANY;
+	ctx->magic_port = INPORT_ANY;
+	len = sizeof(saddr);
+	memset(&saddr, 0, len);
+
+	if(config_bool(NULL, "AllowTransProxy", 0)) {
+		u_int32_t addr = INADDR_ANY;
+		u_int16_t port = INPORT_ANY;
+
+		if(!socket_orgdst(ctx->cli_ctrl, &addr, &port) &&
+		   INADDR_ANY != addr && INADDR_NONE != addr &&
+		   INPORT_ANY != port)
+		{
+			/*
+			** check if destination is a local IP;
+			** ignore to loop-protection...
+			*/
+			int rc = socket_chkladdr(addr);
+
+			switch( rc) {
+			case -1:
+				syslog_write(T_ERR,
+				"check of transparent proxy dest failed");
+			break;
+			case 0:
+				ctx->magic_addr = ntohl(addr);
+				ctx->magic_port = ntohs(port);
+				syslog_write(T_INF,
+				"transparent proxy request to %s:%d from %s",
+				socket_addr2str(ctx->magic_addr),
+				ctx->magic_port, ctx->cli_ctrl->peer);
+			break;
+			default:
+				syslog_write(T_DBG,
+				"requested transparent proxy dest %s is local",
+				socket_addr2str(ntohl(addr)));
+			break;
+			}
+		}
+	}
+
 
 	/*
 	** Check for permission and existence of "magic"
 	** destination address and port information
 	*/
-	ctx->magic_addr = 0;
-	ctx->magic_port = 0;
-	if (config_bool(NULL, "AllowMagicUser", 0) != 0 &&
-				(p = strchr(arg, '@')) != NULL) {
+	if(config_bool(NULL, "AllowMagicUser", 0) != 0 &&
+           (p = config_str(NULL, "UseMagicChar", "@")) &&
+	   (p = strchr(arg, *p)) != NULL)
+	{
 		*p++ = '\0';
 		if ((q = strchr(p, ':')) != NULL) {
 			*q++ = '\0';
@@ -402,7 +457,7 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 			ctx->magic_port = IPPORT_FTP;
 		ctx->magic_addr = socket_str2addr(p, INADDR_ANY);
 		if (ctx->magic_addr == INADDR_ANY ||
-				ctx->magic_port == (u_int16_t) 0) {
+		    ctx->magic_port == INPORT_ANY) {
 			client_respond(501, NULL,
 				"Invalid destination in user name");
 			syslog_write(U_WRN,
@@ -415,81 +470,168 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 	/*
 	** Retrieve the relevant user information
 	*/
-	ctx->username = misc_strdup(FL, arg);
 	if (ctx->magic_addr != INADDR_ANY) {
 		syslog_write(U_INF, "'USER %s' dest %s:%d from %s",
 				arg, socket_addr2str(ctx->magic_addr),
-				(int) ctx->magic_port,
-				ctx->cli_ctrl->peer);
+				(int)ctx->magic_port, ctx->cli_ctrl->peer);
+	} else
+	if(config_str(NULL, "DestinationAddress", NULL) == NULL) {
+		syslog_write(U_ERR, "unknown destination address");
+		client_respond(501, NULL,"Unknown destination address");
+		return;
 	} else {
 		syslog_write(U_INF, "'USER %s' from %s",
 				arg, ctx->cli_ctrl->peer);
 	}
+	ctx->username = misc_strdup(FL, arg);
 	ldap_setup_user(ctx);
 
+	/*
+	** should we bind a rand(port-range) or increment?
+	*/
+	incr = !config_bool(NULL,"SockBindRand", 0);
+
+	/*
+	** mark socket invalid
+	*/
+	sock = -1;
+
 	/*
 	** Forward connection to destination
 	*/
-	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
-		syslog_error("can't create Srv-Ctrl for %s",
-					ctx->cli_ctrl->peer);
-		exit(EXIT_FAILURE);
-	}
-	socket_opts(sock, SK_CONTROL);
-
-	/*
-	** Bind the socket, taking care of a given port range
-	*/
-	if (ctx->srv_lrng > 0 && ctx->srv_urng > 0) {
-		memset(&saddr, 0, sizeof(saddr));
-		saddr.sin_addr.s_addr = INADDR_ANY;
-		saddr.sin_family      = AF_INET;
-
-		for (lprt = ctx->srv_lrng;
-				lprt <= ctx->srv_urng; lprt++) {
-#if defined(COMPILE_DEBUG)
-			debug(2, "try to con-bind Srv-Ctrl to 0:%d",
-							(int) lprt);
-#endif
-			saddr.sin_port = htons(lprt);
-			if (bind(sock, (struct sockaddr *) &saddr,
-						sizeof(saddr)) == 0)
-				break;
-			if (errno != EADDRINUSE)
-				lprt = ctx->srv_urng + 1;
-		}
-		if (lprt > ctx->srv_urng) {	/* nothing found? */
-			close(sock);
-			syslog_error("can't bind Srv-Ctrl "
-					"to 0:%d-%d for %s",
-					(int) ctx->srv_lrng,
-					(int) ctx->srv_urng,
+	retry = MAX_RETRIES;
+	lprt  = ctx->srv_lrng;
+	while(0 <= retry--) {
+		/*
+		** First of all, get a socket
+		*/
+		if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
+			syslog_error("Srv-Ctrl: can't create socket for %s",
 					ctx->cli_ctrl->peer);
 			exit(EXIT_FAILURE);
 		}
+		socket_opts(sock, SK_CONTROL);
+
+		/*
+		** check if we have to take care to a port range
+		*/
+		if( !(INPORT_ANY == ctx->srv_lrng &&
+                      INPORT_ANY == ctx->srv_urng))
+		{
+			u_int32_t ladr = INADDR_ANY;
+
+			/*
+			** bind the socket, taking care of a given port range
+			*/
+			if(incr) {
+				lowrng = lprt;
+#if defined(COMPILE_DEBUG)
+				debug(2, "Srv-Ctrl: "
+				         "about to bind to %s:range(%d-%d)",
+				         socket_addr2str(ladr),
+				         lowrng, ctx->srv_urng);
+#endif
+				bind_res = socket_d_bind(sock, ladr,
+				       lowrng, ctx->srv_urng, incr);
+			} else {
+				lowrng = ctx->srv_lrng;
+#if defined(COMPILE_DEBUG)
+				debug(2, "Srv-Ctrl: "
+				         "about to bind to %s:range(%d-%d)",
+				         socket_addr2str(ladr),
+				         lowrng, ctx->srv_urng);
+#endif
+				bind_res = socket_d_bind(sock, ladr,
+				           lowrng, ctx->srv_urng, incr);
+			}
+			if (INPORT_ANY == bind_res) {
+				/* nothing found? */
+				close(sock);
+				syslog_error("Srv-Ctrl: can't bind to"
+				             " %s:%d for %s",
+					     socket_addr2str(ladr),
+					     (int)lprt, ctx->cli_ctrl->peer);
+				exit(EXIT_FAILURE);
+			} else {
+				lprt = bind_res;
+			}
+		} else lprt = INPORT_ANY;
+
+		/*
+		** Okay, now try the actual connect to the server
+		*/
+		memset(&saddr, 0, sizeof(saddr));
+		saddr.sin_addr.s_addr = htonl(ctx->srv_addr);
+		saddr.sin_family      = AF_INET;
+		saddr.sin_port        = htons(ctx->srv_port);
+
+		if (connect(sock, (struct sockaddr *)&saddr,
+		            sizeof(saddr)) < 0)
+		{
+#if defined(COMPILE_DEBUG)
+				debug(2, "Srv-Ctrl: connect failed with '%s'",
+				      strerror(errno));
+#endif
+			close(sock);
+			sock = -1;
+			/* check if is makes sense to retry?
+			** perhaps we only need an other
+			** local port (EADDRNOTAVAIL) for
+			** this destination?
+			*/
+			if( !(EINTR == errno ||
+			      EAGAIN == errno ||
+			      EADDRINUSE == errno ||
+			      EADDRNOTAVAIL == errno))
+			{
+				/*
+				** an other (real) error ocurred
+				*/
+				syslog_error("Srv-Ctrl: "
+				             "can't connect %s:%d for %s",
+				             socket_addr2str(ctx->srv_addr),
+				             (int) ctx->srv_port,
+				             ctx->cli_ctrl->peer);
+				exit(EXIT_FAILURE);
+			}
+			if(incr && INPORT_ANY != lprt) {
+				/* increment lower range if we use
+				** increment mode and have a range
+				*/
+				if(lprt < ctx->srv_urng) {
+					lprt++;
+				} else {
+				/*
+				** no more ports in range we can try
+				*/
+				syslog_error("Srv-Ctrl: "
+				             "can't connect %s:%d for %s",
+				             socket_addr2str(ctx->srv_addr),
+				             (int) ctx->srv_port,
+				             ctx->cli_ctrl->peer);
+				exit(EXIT_FAILURE);
+				}
+			}
+		} else break;
 	}
 
 	/*
-	** Okay, now try the actual connect to the server
+	** check if we have a valid, connected socket
 	*/
-	memset(&saddr, 0, sizeof(saddr));
-	saddr.sin_addr.s_addr = htonl(ctx->srv_addr);
-	saddr.sin_family      = AF_INET;
-	saddr.sin_port        = htons(ctx->srv_port);
-
-	if (connect(sock, (struct sockaddr *) &saddr,
-					sizeof(saddr)) < 0) {
-		syslog_error("can't connect %s:%d for %s",
-			socket_addr2str(ctx->srv_addr),
-			(int) ctx->srv_port, ctx->cli_ctrl->peer);
+	if(-1 == sock) {
+		syslog_error("Srv-Ctrl: can't connect %s:%d for %s",
+		             socket_addr2str(ctx->srv_addr),
+		             (int) ctx->srv_port,
+		             ctx->cli_ctrl->peer);
 		exit(EXIT_FAILURE);
 	}
+
 	if ((ctx->srv_ctrl = socket_init(sock)) == NULL)
 		misc_die(FL, "cmds_user: ?srv_ctrl?");
 	ctx->srv_ctrl->ctyp = "Srv-Ctrl";
 
 #if defined(COMPILE_DEBUG)
-	debug(2, "Srv-Ctrl is %s:%d",
+		debug(2, "Srv-Ctrl is %s:%d",
 			ctx->srv_ctrl->peer, (int) ctx->srv_port);
 #endif
 
@@ -647,7 +789,7 @@ static void cmds_port(CONTEXT *ctx, char *arg)
 		syslog_write(U_WRN,
 			"syntax error in 'PORT' from %s",
 			ctx->cli_ctrl->peer);
-		client_data_reset();
+		client_data_reset(MOD_RESET);
 		return;
 	}
 	addr = (h1 << 24) + (h2 << 16) + (h3 << 8) + h4;
@@ -663,7 +805,7 @@ static void cmds_port(CONTEXT *ctx, char *arg)
 		syslog_write(U_WRN,
 			"different address in 'PORT' from %s",
 			ctx->cli_ctrl->peer);
-		client_data_reset();
+		client_data_reset(MOD_RESET);
 		return;
 	}
 
@@ -715,6 +857,7 @@ static void cmds_pasv(CONTEXT *ctx, char *arg)
 	u_int16_t port;
 	char str[1024], *p, *q;
 	FILE *fp;
+	int  incr;
 
 	if (ctx == NULL)		/* Basic sanity check	*/
 		misc_die(FL, "cmds_pasv: ?ctx?");
@@ -731,12 +874,28 @@ static void cmds_pasv(CONTEXT *ctx, char *arg)
 	}
 
 	/*
-	** Open a socket that is good for listening
+	** should we bind a rand(port-range) or increment?
 	*/
-	addr = socket_sck2addr(ctx->cli_ctrl->sock, LOC_END, NULL);
+	incr = !config_bool(NULL,"SockBindRand", 0);
+
+	/*
+	** Open a socket that is good for listening
+	**
+	** TransProxy mode: check if we can use our real
+	** ip instead of the server's one as our local ip,
+	** we bind the socket/ports to.
+	*/
+	addr = INADDR_ANY;
+	if(config_bool(NULL, "AllowTransProxy", 0)) {
+		addr = config_addr(NULL, "Listen", (u_int32_t)INADDR_ANY);
+	}
+	if(INADDR_ANY == addr) {
+		addr = socket_sck2addr(ctx->cli_ctrl->sock, LOC_END, NULL);
+	}
 	if ((port = socket_d_listen(addr, ctx->pas_lrng, ctx->pas_urng,
-				&(ctx->cli_data), "Cli-Data")) == 0) {
-		syslog_error("can't bind Cli-Data to %s:%d-%d for %s",
+			&(ctx->cli_data), "Cli-Data", incr)) == 0)
+	{
+		syslog_error("Cli-Data: can't bind to %s:%d-%d for %s",
 			socket_addr2str(addr), (int) ctx->pas_lrng,
 			(int) ctx->pas_urng, ctx->cli_ctrl->peer);
 		client_respond(425, NULL, "Can't open data connection");
@@ -806,7 +965,7 @@ static void cmds_pasv(CONTEXT *ctx, char *arg)
 
 static void cmds_xfer(CONTEXT *ctx, char *arg)
 {
-	int mode;
+	int mode = MOD_ACT_FTP;
 	char *cmd;
 	u_int32_t addr;
 	u_int16_t port;
@@ -852,12 +1011,18 @@ static void cmds_xfer(CONTEXT *ctx, char *arg)
 	** In active mode we listen and the server connects
 	*/
 	if (mode == MOD_ACT_FTP) {
+		/*
+		** should we bind a rand(port-range) or increment?
+		*/
+		int incr = !config_bool(NULL,"SockBindRand", 0);
+
 		addr = socket_sck2addr(ctx->srv_ctrl->sock,
 						LOC_END, NULL);
+
 		if ((port = socket_d_listen(addr, ctx->srv_lrng,
 				ctx->srv_urng, &(ctx->srv_data),
-				"Srv-Data")) == 0) {
-			syslog_error("can't bind Srv-Data to "
+				"Srv-Data", incr)) == 0) {
+			syslog_error("Srv-Data: can't bind to "
 					"%s:%d-%d for %s",
 					socket_addr2str(addr),
 					(int) ctx->srv_lrng,
@@ -865,7 +1030,7 @@ static void cmds_xfer(CONTEXT *ctx, char *arg)
 					ctx->cli_ctrl->peer);
 			client_respond(425, NULL,
 					"Can't open data connection");
-			client_data_reset();
+			client_data_reset(MOD_RESET);
 			return;
 		}
 
@@ -921,7 +1086,7 @@ static void cmds_abor(CONTEXT *ctx, char *arg)
 	/*
 	** Reset data connection variables (esp. PASV)
 	*/
-	client_data_reset();
+	client_data_reset(MOD_RESET);
 
 	/*
 	** If no transfer is in progress, don't worry
@@ -1036,6 +1201,37 @@ static void cmds_epsv(CONTEXT *ctx, char *arg)
 #endif
 
 
+#if defined(ENABLE_SSL) /* <!-- SSL --> */
+/* ------------------------------------------------------------ **
+**
+**	Function......:	cmds_auth
+**
+**	Parameters....:	ctx		Pointer to user context
+**			arg		Command argument(s)
+**
+**	Return........:	(none)
+**
+**	Purpose.......: Act upon the 'AUTH' command.
+**			This is for SSL authentication.
+**
+** ------------------------------------------------------------ */
+
+static void cmds_auth(CONTEXT *ctx, char *arg)
+{
+	if (ctx == NULL)		/* Basic sanity check	*/
+		misc_die(FL, "cmds_auth: ?ctx?");
+
+	if (arg == NULL || strcasecmp(arg, "SSL") != 0) {
+		client_respond(501, NULL,
+				"Missing or bad auth method");
+		return;
+	}
+
+	/* TODO */
+}
+
+
+#endif /* <!-- /SSL --> */
 #if defined(HAVE_REGEX)
 /* ------------------------------------------------------------ **
 **
@@ -1087,8 +1283,12 @@ char *cmds_reg_comp(void **ppre, char *ptr)
 			str[i++] = *ptr;
 			continue;
 		}
-		if (isxdigit(ptr[1]) && isxdigit(ptr[2])) {
+		if (isxdigit((int)ptr[1]) && isxdigit((int)ptr[2])) {
+#if defined(HAVE_SNPRINTF)
+			snprintf(tmp, sizeof(tmp), "%.2s", ptr + 1);
+#else
 			sprintf(tmp, "%.2s", ptr + 1);
+#endif
 			sscanf(tmp, "%x", &c);
 			str[i++] = (char) c;
 			ptr += 2;
@@ -1157,6 +1357,14 @@ char *cmds_reg_exec(void *regex, char *str)
 
 /* ------------------------------------------------------------
  * $Log: ftp-cmds.c,v $
+ * Revision 1.9  2002/01/14 19:39:30  mt
+ * implemented workarround for Netscape (4.x) directory symlink handling
+ * changed to socket_orgdst usage to get ransparent proxy destination
+ *
+ * Revision 1.8  2001/11/06 23:04:44  mt
+ * applied / merged with transparent proxy patches v8
+ * see ftp-proxy/NEWS for more detailed release news
+ *
  * Revision 1.7  1999/10/19 10:19:15  wiegand
  * use port range also for control connection to server
  *
